/*! 
 * @file fi_hwmon.c
 * @brief Hardware monitor interface for the Fastec Imaging FPGA driver.
 *
 * Copyright (C) 2021 Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------

#include <linux/bits.h>
#include <linux/clk.h>
#include <linux/delay.h>
#include <linux/hwmon.h>
#include <linux/io.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/of.h>

#include "fi_pci.h"
#include "fi_fpga_dev.h"

//-----------------------------------------------------------------------------
// Constant Definitions
//-----------------------------------------------------------------------------

//! The name we call ourself. 
#define HWMON_NAME 		"fi_hwmon"

const uint8_t EMC2113_I2C_ADDR = 0x2E;      //!< I2C address of the fan controller on the IO board.

const uint8_t EMC2113_EXT_DIODE1_HIGH = 0x02;
const uint8_t EMC2113_EXT_DIODE1_LOW = 0x03;

const uint8_t EMC2113_PWM_VALUE = 0x40;

const uint8_t EMC2113_TACH_VALUE_HIGH = 0x4E;
const uint8_t EMC2113_TACH_VALUE_LOW = 0x4F;

//-----------------------------------------------------------------------------
// Variable Declarations
//-----------------------------------------------------------------------------

// ===================================
// TODO:
// - The PWM isn't very interesting until you allow the system to set the PWM.
// - Other temperature sensors on the head should be added.
// ===================================

static const struct hwmon_channel_info *fi_hwmon_info[] = {
        HWMON_CHANNEL_INFO(chip,
                        HWMON_C_REGISTER_TZ | HWMON_C_UPDATE_INTERVAL),
        HWMON_CHANNEL_INFO(temp,
                        HWMON_T_INPUT | HWMON_T_MAX | HWMON_T_MAX_HYST),
	HWMON_CHANNEL_INFO(pwm,
			HWMON_PWM_INPUT),
	HWMON_CHANNEL_INFO(fan,
			HWMON_F_INPUT),
        NULL
};

static bool i2c_wait_idle(uint8_t *data_ptr, bool two_bytes)
{
    int	retry_count;
    int max_retries = 10;
    uint32_t i2c_status = I2CM_STATUS_BUSY;
    extern FI_FPGA_DEV fi_fpga_dev;

//    printk("%s\n", __FUNCTION__);

    // Control access to the driver.
    for (retry_count = 0; retry_count < max_retries; ++retry_count)
    {
	// Check to see if the busy bit is still set.
	i2c_status = fi_get_fpga_reg(&fi_fpga_dev, I2C_MASTER_STATUS_IOB);
        if ((i2c_status & I2CM_STATUS_BUSY) == 0)
	    break;

	// Wait for a little bit.
	usleep_range(500, 2000);
    }

    if (i2c_status & I2CM_STATUS_BUSY)
    {
	printk("%s: timeout: 0x%08x\n", __FUNCTION__, i2c_status);
    }
    else if ((i2c_status & I2CM_STATUS_ERR) != 0)
    {
	printk("%s: error bit set: 0x%08x\n", __FUNCTION__, i2c_status);
    }
    else if (data_ptr != NULL)
    {
	uint32_t data_val;

	// Read the data and control registers.
	data_val = fi_get_fpga_reg(&fi_fpga_dev, I2C_MASTER_RD_DATA_IOB);

	// Do we have 1 or 2 bytes?
	if (two_bytes)
	{
	    *data_ptr = ((data_val >> 8) & 0xFF);
	    ++data_ptr;
	}
	
        *data_ptr = (data_val & 0xFF);
    }

    // Did we exceed our retries?
    return ((i2c_status & I2CM_STATUS_BUSY) == 0);
}

bool i2c_read(uint8_t chip_addr, uint8_t reg_addr, uint8_t *reg_data, bool two_bytes)
{
    uint32_t	reg_value;
    extern FI_FPGA_DEV fi_fpga_dev;

//    printk("%s\n", __FUNCTION__);

    // Make sure the I2C bus is available.
    if (!i2c_wait_idle(NULL, false))
    {
        printk(KERN_ERR "%s: I2C never becomes idle.\n", __FUNCTION__ );
	return false;
    }

    // Specify the register address we're going to read from.
    fi_set_fpga_reg(&fi_fpga_dev, I2C_MASTER_CMD_DATA_IOB, (uint32_t)reg_addr);

    // Read the data.
    reg_value = I2CM_CMD_READ;
    reg_value |= I2CM_CMD_DEV(chip_addr);
    reg_value |= I2CM_CMD_1CMD_CNT;
    if (two_bytes)
	reg_value |= I2CM_CMD_2DATA_CNT;
    else 
	reg_value |= I2CM_CMD_1DATA_CNT;
    fi_set_fpga_reg(&fi_fpga_dev, I2C_MASTER_CONTROL_IOB, reg_value);

    // Now wait for the data to come back.
    if (!i2c_wait_idle(reg_data, two_bytes))
    {
        printk(KERN_ERR "%s: I2C never becomes idle.\n", __FUNCTION__ );
	return false;
    }
    
#if 0
    if (two_bytes)
	printk("[%02X] = 0x%02X%02X\n", reg_addr, reg_data[1], reg_data[0]);
    else
	printk("[%02X] = 0x%02X\n", reg_addr, reg_data[0]);
#endif

    // Return our status.
    return true;
}



/*!
 * @brief Return the fan RPM.
 */
uint32_t read_fan_tach(void)
{
    uint8_t tach[2];
    uint32_t fan_tach;

//    printk("%s\n", __FUNCTION__);

    // Read the high and low bytes for the external diode1.
    tach[0] = tach[1] = 0;
    if (!i2c_read(EMC2113_I2C_ADDR, EMC2113_TACH_VALUE_HIGH, tach, true))
    {
        printk(KERN_ERR "%s: Error doing I2C!\n", __FUNCTION__);
        return 0;
    }

    // Convert to the tach value. From the datasheet...
    fan_tach = ((((uint32_t)tach[0]) << 5) | ((uint32_t)tach[1] >> 3));
    fan_tach = (3932160 * 2) / fan_tach;


    // All done. Need things in units of millidegrees Celsius.
    return fan_tach;
}

/*!
 * @brief Return the fan PWM.
 */
uint32_t read_fan_pwm(void)
{
    uint8_t pwm_value;

//    printk("%s\n", __FUNCTION__);

    // Read the high and low bytes for the external diode1.
    pwm_value = 0;
    if (!i2c_read(EMC2113_I2C_ADDR, EMC2113_PWM_VALUE, &pwm_value, false))
    {
        printk(KERN_ERR "%s: Error doing I2C!\n", __FUNCTION__);
        return 0;
    }

    // All done. Need things in units of millidegrees Celsius.
    return (uint32_t)pwm_value;
}

/*!
 * @brief Return the fan temperature in millidegrees Celsius.
 */
uint32_t read_fan_temp(void)
{
    uint8_t temp[2];
    int32_t degrees_c;

//    printk("%s\n", __FUNCTION__);

    // Read the high and low bytes for the external diode1.
    temp[0] = temp[1] = 0;
    if (!i2c_read(EMC2113_I2C_ADDR, EMC2113_EXT_DIODE1_HIGH, &temp[0], false))
    {
        printk(KERN_ERR "%s: Error doing I2C!\n", __FUNCTION__);
        return 0;
    }
    if (!i2c_read(EMC2113_I2C_ADDR, EMC2113_EXT_DIODE1_LOW, &temp[1], false))
    {
        printk(KERN_ERR "%s: Error doing I2C!\n", __FUNCTION__);
        return 0;
    }

    // Convert to the full temperature. To get things in millidegrees:
    //    temp = (((T0 * 256) + T1) * 1000) / 256
    degrees_c = (((uint32_t)(temp[0] & 0x7F)) << 8) + ((uint32_t)temp[1]);
    if ((temp[0] & 0x80) == 0x80)
        degrees_c = -degrees_c;

    // All done. Need things in units of millidegrees Celsius.
    return (uint32_t)((degrees_c * 1000) / 256);
}


static int fi_hwmon_read(struct device *dev,
			 enum hwmon_sensor_types type,
			 u32 attr, int channel, long *val)
{
    int rc;

//    printk("%s: attr=%u, channel=%d\n", __FUNCTION__, attr, channel);

    // Initialize to an error condition.
    rc = -ENOTSUPP;

    switch (type)
    {
    case hwmon_fan:
	if (attr == hwmon_fan_input)
	{
	    // Read the tach from the fan.
	    *val = read_fan_tach();
	    rc = 0;
	}
	break;
    case hwmon_pwm:
	if (attr == hwmon_pwm_input)
	{
	    // Read the PWM from the fan.
	    *val = read_fan_pwm();
	    rc = 0;
	}
	break;
    case hwmon_temp:
	if (attr == hwmon_temp_input)
	{
	    // Read the temperature from the fan.
	    *val = read_fan_temp();
	    rc = 0;
	}
	break;
    default:
	break;
    }

    // Return our result.
    return rc;
}

static int fi_hwmon_write(struct device *dev,
			  enum hwmon_sensor_types type,
			  u32 attr, int channel, long val)
{
    int rc;

//    printk("%s\n", __FUNCTION__);

    // Initialize to an error condition.
    rc = -ENOTSUPP;

    switch (type)
    {
    case hwmon_pwm:
    default:
	break;
    }

    // Return our result.
    return rc;
}



/*!
 * @brief Called to determine if a particular entry should be visible in the
 * 	  file system or not.
 */
umode_t fi_hwmon_is_visible(const void *data, enum hwmon_sensor_types type, u32 attr, int channel)
{
    umode_t rc;

//    printk("%s\n", __FUNCTION__);

    // Initialize our return value to NOT visible.
    rc = 0;

    // Which sensor are we talking about?
    switch (type)
    {
    case hwmon_fan:
	switch (attr)
	{
	case hwmon_fan_input:
	case hwmon_fan_label:
	    rc = 0444;
	    break;
	default:
	    break;
	}
	break;
    case hwmon_pwm:
	switch (attr)
	{
	case hwmon_pwm_input:
	    rc = 0444;
	    break;
	default:
	    break;
	}
	break;
    case hwmon_temp:
	switch (attr)
	{
	case hwmon_temp_input:
	    rc = 0444;
	    break;
	default:
	    break;
	}
	break;
    default:
	break;
    }

    // Return the mode.
    return rc;
}

static int fi_hwmon_read_labels(struct device *dev,
			        enum hwmon_sensor_types type,
			        u32 attr, int channel, const char **str)
{
    int rc;

//    printk("%s\n", __FUNCTION__);

    // Initialize to an error condition.
    rc = -ENOTSUPP;

    switch (type)
    {
    case hwmon_fan:
	*str = "FAN";
	rc = 0;
	break;
    case hwmon_pwm:
	*str = "PWM";
	rc = 0;
	break;
    case hwmon_temp:
	*str = "FTEMP";
	rc = 0;
	break;
    default:
	break;
    }

    // Return our result.
    return rc;
}

static const struct hwmon_ops fi_hwmon_ops = {
        .is_visible = fi_hwmon_is_visible,
        .read = fi_hwmon_read,
        .write = fi_hwmon_write,
	.read_string = fi_hwmon_read_labels,
};

static const struct hwmon_chip_info fi_hwmon_chip_info = {
        .ops = &fi_hwmon_ops,
        .info = fi_hwmon_info,
};

/*! 
 * @brief Set up our hwmon device.
 */
int fi_fpga_setup_hdev(void)
{
    extern FI_FPGA_DEV	fi_fpga_dev;

//    printk("%s\n", __FUNCTION__);

    fi_fpga_dev.hdev = devm_hwmon_device_register_with_info(fi_fpga_dev.fpga_dev,
						     HWMON_NAME,
						     &fi_fpga_dev,
						     &fi_hwmon_chip_info,
						     NULL);

    return PTR_ERR_OR_ZERO(fi_fpga_dev.hdev);
}

