/*! 
 * @file fpga_fops.c
 * @brief File operation functions for the FPGA device driver.
 *
 * Copyright (C) 2010,2015 Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------

#include <linux/device.h>
#include <linux/delay.h>
#include <linux/types.h>	/* size_t */
#include <linux/errno.h>	/* error codes */
#include <linux/fs.h>		/* everything... */
#include <linux/slab.h>
#include <linux/uaccess.h>
#include <linux/mutex.h>
#include <linux/jiffies.h>
// Old version: 
#include <linux/sched/signal.h>
// Instead use:
#include <linux/sched.h>
#include <linux/wait.h>
#include <asm/fpu/api.h>

#include "fi_pci.h"
#include "fi_fpga_dev.h"

//-----------------------------------------------------------------------------
// Constant Definitions
//-----------------------------------------------------------------------------

//! How many bytes in the IPM histogram.
#define IPM_HISTOGRAM_SIZE	0xC00	// 256 4 byte entries each for red/green/blue

// Define the maximum number of events we'll buffer.
#define MAX_EVENT_BUFS		16

// Number of milliseconds to wait for a PCIE2IM DMA operation to complete.
#define PCIE2IM_XFER_TIMEOUT		5000

// Number of milliseconds to wait for an IM2PCIE DMA operation to complete.
#define IM2PCIE_XFER_TIMEOUT		5000

// Define a macro to increment the ring buffer indicies.
#define EVBUF_INCR( x )		(((x)+1)&(MAX_EVENT_BUFS-1))

// Define the events we are interested in.
#define INTERESTING_EVENTS	(PCIEIRQ_FIFO_NOT_EMPTY | \
				 PCIEIRQ_P1_FRAME_DONE  | \
				 PCIEIRQ_EXT_ARM_INPUT | \
				 PCIEIRQ_EXT_UNARM_INPUT | \
				 PCIEIRQ_LR_DESC_FIFO_AF | \
				 PCIEIRQ_PWR_OFF  | \
				 PCIEIRQ_MFT_LSDET | PCIEIRQ_MFT_LSLPSW | \
				 PCIEIRQ_TEST)

//-----------------------------------------------------------------------------
// Variable Declarations
//-----------------------------------------------------------------------------

// Declare and initialize our mutex on image memory accesses.
static DEFINE_MUTEX(im_lock);

// Declare a wait queue for when we're waiting on the IM2PCIE or PCIE2IM DMA.
static DECLARE_WAIT_QUEUE_HEAD(xfer_frame_q);
static int frame_event_value = 0;

// Declare a wait queue for when we're waiting on events from the FPGA.
static DECLARE_WAIT_QUEUE_HEAD(event_q);
static uint32_t event_value;

static uint32_t	ev_buf[ MAX_EVENT_BUFS ];	//!< The ring of event buffers.
static atomic_t	evbuf_rd;			//!< The ring buffer read index.
static atomic_t	evbuf_wr;			//!< The ring buffer write index.

//-----------------------------------------------------------------------------
// Function Prototypes
//-----------------------------------------------------------------------------

static void dma_done_irq(u32 irq_status);
static void ev_callback(u32 irq_status);
static int get_event( struct file *filp, char __user *buf );
static int read_fpga_reg( FI_FPGA_DEV *dev, FI_REG_ACCESS *reg_access );
static int write_fpga_reg( FI_FPGA_DEV *dev, FI_REG_ACCESS *reg_access );
static int read_ipm_histogram( FI_FPGA_DEV *dev, uint32_t __user *arg );


/*!
 * @brief Called by the IRQ handler when the interrupt is one of our 
 *        registered events.
 */
static void ev_callback(u32 irq_status)
{
    int rd_index = atomic_read(&evbuf_rd);
    int	wr_index = atomic_read(&evbuf_wr);

//    printk("%s: irq_status=0x%08x\n", __FUNCTION__, irq_status);

    // Make sure we aren't full. Full is when the write index
    // is one behind the read index.
    if (rd_index == EVBUF_INCR(wr_index))
    {
	// The event buffer is full.
	printk( KERN_ERR "%s: Event buffer is FULL!\n", __FUNCTION__ );
    }
    else
    {
	// Save the event information.
	ev_buf[wr_index] = irq_status;
	atomic_set(&evbuf_wr, EVBUF_INCR(wr_index));
    }

    // Save the status for the event.
    event_value = irq_status;

    // Wake up anyone who might be waiting.
    wake_up_interruptible(&event_q);
}

static int read_im_fifo( FI_FPGA_DEV *dev, uint32_t __user *arg )
{
    int ret_value;
    uint32_t *pci_im_fifo;

    // Get the address of the IM FIFO.
    pci_im_fifo = fi_pci_get_im_fifo();

#ifdef USE_IOMEM
    // Copy data from the image fifo to our local buffer.
    memcpy_fromio(dev->temp_buffer, pci_im_fifo, 16);
#else
    // ------------------------------------------------------
    // This is WAY more efficient than the memcpy...but it
    // broke as of 4.16.10. Suddenly the pci_im_fifo wasn't mapping
    // as a virtual address.
    // ------------------------------------------------------
    //
    // Read 16 bytes.
    kernel_fpu_begin();
    asm("movntdqa (%0), %%xmm0\n"
	"movaps %%xmm0, (%1)\n"
	:: "r" (pci_im_fifo), "r" (dev->temp_buffer) : "memory");
    kernel_fpu_end();
#endif

    // Copy the data to the user.
    if (( ret_value = copy_to_user( arg, dev->temp_buffer, 16 )))
    {
	printk( KERN_ERR "%s: copy_to_user() failed!", __FUNCTION__ );
	ret_value = -EFAULT;
    }

    // Return the result.
    return ret_value;
}

/*! 
 * @brief Read the histogram (presumably previously captured) from the Image Processing
 * 	  Module (IPM) of the FPGA.
 * 
 * @param dev	Pointer to our device structure.
 * @param arg	Pointer to the user buffer to receive the data.
 * 
 * @return 0 on success, <0 otherwise.
 */
static int read_ipm_histogram( FI_FPGA_DEV *dev, uint32_t __user *arg )
{
    int	count;
    int ret_value;
    uint32_t	temp_value;
    uint32_t	*temp_ptr;

    PDEBUG( 2, "%s: dev=%p\n", __FUNCTION__, dev );

    // Read the status register and make sure the histogram is ready.
    temp_value = fi_get_fpga_reg( dev, IPM_STATUS );
    PDEBUG( 3, "%s: IPM_STATUS=0x%x\n", __FUNCTION__, temp_value );
    if ( temp_value & IPMS_HISTO_BUSY )
    {
	printk( "%s: Histogram is not ready.\n", __FUNCTION__ );
	return -EBUSY;
    }

    // Pull over the histogram data.
    temp_ptr = dev->temp_buffer;
    for ( count = 0; count < IPM_HISTOGRAM_SIZE; count += sizeof( uint32_t ) )
	*temp_ptr++ = fi_get_fpga_reg( dev, IPM_RGB_HISTO_TABLE_RED_START + count );

    // Copy the data to the user.
    if (( ret_value = copy_to_user( arg, dev->temp_buffer, IPM_HISTOGRAM_SIZE )))
    {
	printk( KERN_ERR "%s: copy_to_user() failed!", __FUNCTION__ );
	ret_value = -EFAULT;
    }

    // Return our status.
    return ret_value;
}

/*! 
 * @brief Get an event from the FPGA
 * 
 * Applications call the FI_FPGA_IOC_GETEVENT ioctl to wait for an event
 * from the FPGA. When an interrupt happens, this returns to the user
 * with the value from the interrupt source register.
 *
 * By default FI_FPGA_IOC_GETEVENT blocks; when the O_NONBLOCK flag
 * was given to the open() function, we return immediately with whatever
 * value is in the interrupt source register.
 *
 * If the user specifies a timeout value, we'll wait for that period.
 *
 * @param filep	Pointer to our file structure.
 * @param buf	Pointer to the user buffer to receive the event information.
 * 
 * @return 0 on success, <0 otherwise.
 */
static int get_event( struct file *filp, char __user *buf )
{
    FI_FPGA_DEV	  *dev;
    FI_EVENT_INFO event_info;
    uint32_t wait_jiffies;
    int	wait_status;
    int	ret_value;
    int	rd_index;
    int wr_index;

    PDEBUG( 2, "%s\n", __FUNCTION__ );

    // Get our pointers.
    dev = (FI_FPGA_DEV *)filp->private_data;
    if ( dev == NULL )
    {
	printk( KERN_ERR "%s - No user data?!\n", __FUNCTION__ );
	return -EFAULT;
    }

    // Grab the structure from the user.
    if ( copy_from_user( &event_info, buf, sizeof( FI_EVENT_INFO )))
    {
	printk( KERN_ERR "%s: Unable to read event info from user.", __FUNCTION__ );
	return -EFAULT;
    }

    // Can we block?
    if (( filp->f_flags & O_NONBLOCK ) == 0 )
    {
	// Initialize our event condition.
	event_value = 0;

	// Get our read/write indicies.
	rd_index = atomic_read(&evbuf_rd);
	wr_index = atomic_read(&evbuf_wr);

	// Do we have a queued event?
	if ( rd_index != wr_index )
	{
	    // Return the accumulated value.
	    event_value = ev_buf[ rd_index ];
	    atomic_set(&evbuf_rd, EVBUF_INCR( rd_index ));
	}
	
	// Did we have an event ready?
	if ( event_value )
	{
	    PDEBUG( 5, "%s[0x%lx]: EARLY status=0x%x\n", __FUNCTION__, jiffies, event_value );
	    goto ge_exit;
	}

	// Is there a wait period?
	if ( event_info.event_timeout != 0 )
	{
	    PDEBUG( 2, "%s: Timeout wait.\n", __FUNCTION__ );

	    // Figure out how many jiffies we have to wait for.
	    wait_jiffies = ( event_info.event_timeout * HZ ) / 1000;

	    // Time to wait.
	    wait_status = wait_event_interruptible_timeout(event_q, (event_value != 0), wait_jiffies );
	    if ( wait_status == 0 )
	    {
		PDEBUG( 2, "%s: TIMEOUT!\n", __FUNCTION__ );
		return -ETIME;
	    }
	    else if ( signal_pending( current ))
	    {
		PDEBUG( 2, "%s: interrupt received\n", __FUNCTION__ );
		return -ERESTARTSYS;
	    }
	}
	else
	{
	    // Wait for an event to roll around.
	    PDEBUG( 5, "%s: Before the wait, dev=0x%p\n", __FUNCTION__, dev );
	    wait_status = wait_event_interruptible(event_q, (event_value != 0));
	    PDEBUG( 5, "%s: After the wait, event_value=0x%x\n", __FUNCTION__, event_value );
	    if ( wait_status )
	    {
		PDEBUG( 1, "%s: Wait interrupted\n", __FUNCTION__ );
		return -ERESTARTSYS;
	    }
	}

	// At this point, we got an event so move the value over.
	rd_index = atomic_read(&evbuf_rd);
	event_value = ev_buf[ rd_index ];
	atomic_set(&evbuf_rd, EVBUF_INCR( rd_index ));
	PDEBUG( 5, "%s[0x%lx]: ret status=0x%x\n", __FUNCTION__, jiffies, event_value );
    }
    else
    {
	PDEBUG( 2, "%s: Non-blocking.\n", __FUNCTION__ );

	// Just read the value straight away.
	event_value = fi_get_fpga_reg( dev, ICM_EVENT_FIFO );
    }

ge_exit:
    // Copy the data back to user space.
    event_info.event_value = event_value;
    if (( ret_value = copy_to_user( buf, &event_info, sizeof( event_info ))))
	printk( KERN_ERR "%s: copy_to_user() failed!", __FUNCTION__ );

    // Return our status.
    return ret_value;
}

/*! 
 * @brief Read an FPGA register.
 * 
 * @param dev Pointer to the FPGA device driver structure.
 * @param reg_access Pointer to the register access structure.
 * 
 * @return 0 if all is well, <0 on error.
 */
static int read_fpga_reg( FI_FPGA_DEV *dev, FI_REG_ACCESS *reg_access )
{
    int 	retval = 0;

    PDEBUG(5, "%s: id=0x%x\n", __FUNCTION__, reg_access->reg_id );

    // Make sure the register ID makes sense.
    if ( !FI_FPGA_REG_VALID( reg_access->reg_id ))
    {
	dev_err( dev->fpga_dev, "Bad read reg 0x%08x\n", reg_access->reg_id );
	return -EFAULT;
    }

    // Grab the info from the FPGA.
    reg_access->value = fi_get_fpga_reg( dev, reg_access->reg_id );

    PDEBUG(5, "%s: value=0x%x [%d]\n", __FUNCTION__, reg_access->value, retval );

    // Return our status.
    return retval;
}

/*! 
 * @brief Write a value to an FPGA register.
 * 
 * @param dev Pointer to the FPGA device driver structure.
 * @param reg_access Pointer to the register access structure.
 * 
 * @return 0 if all is well, <0 on error.
 */
static int write_fpga_reg( FI_FPGA_DEV *dev, FI_REG_ACCESS *reg_access )
{
    int 	retval = 0;

    PDEBUG(5, "%s: id=0x%x, value=0x%x\n", __FUNCTION__, reg_access->reg_id, reg_access->value );

    // Make sure the register ID makes sense.
    if ( !FI_FPGA_REG_VALID( reg_access->reg_id ))
    {
	dev_err( dev->fpga_dev, "Bad write reg 0x%08x\n", reg_access->reg_id );
	return -EFAULT;
    }

    // Update the register on the FPGA.
    fi_set_fpga_reg( dev, reg_access->reg_id, reg_access->value );

    // Return our status.
    return retval;
}

/*! 
 * @brief Called when the user opens the device.
 */
int fi_fpga_open( struct inode *inode, struct file *filp )
{
    FI_FPGA_DEV *dev;		// Pointer to our device structure.

    PDEBUG(2, "%s\n", __FUNCTION__ );

    // Get a pointer to our device structure.
    dev = container_of( inode->i_cdev, FI_FPGA_DEV, fpga_cdev );
    if ( dev == NULL )
    {
	printk( KERN_ERR "%s - NULL device!\n", __FUNCTION__ );
	return -ENODEV;
    }
    filp->private_data = dev;

    // Make sure we don't have too many users. The current design
    // only allows for one user of the driver.
    // The test decrements by one and returns true if 0, false otherwise.
    if (!atomic_dec_and_test(&dev->available))
    {
	// Restore the count.
	atomic_inc(&dev->available);

	// Notify the user there was a problem.
	dev_err( dev->fpga_dev, "fi_fpga_open: Too many users!\n" );
	return -EBUSY;
    }

    // Allocate our temporary buffer.
    dev->temp_buffer = kmalloc(4096, GFP_KERNEL);

    // Initialize our event ring. When the read index is the same as
    // the write index, the buffer is empty.
    atomic_set(&evbuf_rd, 0);
    atomic_set(&evbuf_wr, 0);

    // Register for event callback.
    fi_pci_add_callback(ev_callback, INTERESTING_EVENTS);

    // Register for catching the DMA interrupts.
    fi_pci_add_callback(dma_done_irq, (PCIEIRQ_IM2PCIE_DMA_DONE | PCIEIRQ_PCIE2IM_DMA_DONE));

    // All done.
    return 0;
}

/*! 
 * @brief Called when the user closes the device.
 */
int fi_fpga_release( struct inode *inode, struct file *filp )
{
    FI_FPGA_DEV *dev;		// Pointer to our device structure.

    PDEBUG(2, "%s\n", __FUNCTION__ );

    // Get a pointer to our device structure.
    dev = container_of( inode->i_cdev, FI_FPGA_DEV, fpga_cdev );
    if ( dev == NULL )
    {
	printk( KERN_ERR "%s - NULL device!\n", __FUNCTION__ );
	return -ENODEV;
    }

    // Remove our callbacks.
    fi_pci_remove_callback(ev_callback);
    fi_pci_remove_callback(dma_done_irq);

    // Release our buffer.
    kfree(dev->temp_buffer);
    dev->temp_buffer = NULL;

    // Update our count.
    atomic_inc(&dev->available);

    // All done.
    return 0;
}

/*
 * The ioctl() implementation
 */

long fi_fpga_ioctl( struct file *filp, unsigned int cmd, unsigned long arg )
{
    int 		retval = 0;
    FI_FPGA_DEV	 	*dev;
    FI_REG_ACCESS	reg_access;
    FI_REG_MODIFY	reg_modify;
    FI_IMAGE_ACCESS	im_access;

    PDEBUG( 2, "%s 0x%x\n", __FUNCTION__, cmd );

    // Extract the type and number bitfields, and don't decode
    // wrong cmds: return ENOTTY (inappropriate ioctl) before processing.
    if (_IOC_TYPE(cmd) != FI_FPGA_IOC_MAGIC)
       	return -ENOTTY;
    if (_IOC_NR(cmd) > FI_FPGA_IOC_MAXNR)
       	return -ENOTTY;

    // Get our pointers.
    dev = filp->private_data;
    if ( dev == NULL )
    {
	dev_err( dev->fpga_dev, "ioctl: no device data\n" );
	return -EFAULT;
    }

    // Process the specific command.
    switch ( cmd )
    {
    case FI_FPGA_IOC_RESET:
	retval = 0;
	break;
    case FI_FPGA_IOC_SETREG:
	// Load up our local variable from the user supplied structure.
	retval = copy_from_user( &reg_access, (char __user *)arg, sizeof( FI_REG_ACCESS ));
	if ( retval == 0 )
	{
	    // Now we have the info we need, go ahead and push the data out.
	    retval = write_fpga_reg( dev, &reg_access );
	}
	break;
    case FI_FPGA_IOC_QUERYREG:
	// Load up our local variable from the user supplied structure.
	retval = copy_from_user( &reg_access, (char __user *)arg, sizeof( FI_REG_ACCESS ));
	if ( retval == 0 )
	{
	    // Now we have the info we need, go ahead and read in the data.
	    retval = read_fpga_reg( dev, &reg_access );
	    if ( retval == 0 )
		retval = reg_access.value;
	}
	break;
    case FI_FPGA_IOC_GETREG:
	// Load up our local variable from the user supplied structure.
	retval = copy_from_user( &reg_access, (char __user *)arg, sizeof( FI_REG_ACCESS ));
	if ( retval == 0 )
	{
	    // Read the register.
	    retval = read_fpga_reg( dev, &reg_access );

	    // Copy the contents back to the user.
	    if ( retval == 0 )
		retval = copy_to_user((char __user *)arg, &reg_access, sizeof( FI_REG_ACCESS ));
	}
	break;
    case FI_FPGA_IOC_MODREG:
	// Load up our local variable from the user supplied structure.
	retval = copy_from_user( &reg_modify, (char __user *)arg, sizeof( FI_REG_MODIFY ));
	if ( retval == 0 )
	{
	    // Modify (read, mask, or) the register.
	    reg_modify.value = fi_mod_fpga_reg( dev, reg_modify.reg_id, reg_modify.value, reg_modify.mask );

	    // Copy the contents back to the user.
	    retval = copy_to_user((char __user *)arg, &reg_modify, sizeof( FI_REG_MODIFY ));
	}
	break;
    case FI_FPGA_IOC_READ_IMAGE:
	// Load up our local variable from the user supplied structure.
	retval = copy_from_user( &im_access, (char __user *)arg, sizeof( FI_IMAGE_ACCESS ));
	if ( retval == 0 )
	{
	    // Read the image as per the user's instructions.
	    retval = fi_fpga_image_read( &im_access );
	}
	break;
    case FI_FPGA_IOC_WRITE_IMAGE:
	// Load up our local variable from the user supplied structure.
	retval = copy_from_user( &im_access, (char __user *)arg, sizeof( FI_IMAGE_ACCESS ));
	if ( retval == 0 )
	{
	    // Write the image as per the user's instructions.
	    retval = fi_fpga_image_write( &im_access );
	}
	break;
    case FI_FPGA_IOC_GETEVENT:
	retval = get_event( filp, (char __user *)arg );
	break;
    case FI_FPGA_IOC_GET_IPM_HISTO:
	retval = read_ipm_histogram( dev, (uint32_t __user *)arg );
	break;
    case FI_FPGA_IOC_READ_IM_FIFO:
	retval = read_im_fifo( dev,  (uint32_t __user *)arg );
	break;
    default:  /* redundant, as cmd was checked against MAXNR */
	retval = -ENOTTY;
	break;
    }

    // Return our status and/or result.
    return retval;
}

static void dma_done_irq(u32 irq_status)
{
//    printk("dma_done_irq: irq_status=0x%x", irq_status);

    frame_event_value = irq_status;
    wake_up_interruptible(&xfer_frame_q);
}

int fi_fpga_image_read( FI_IMAGE_ACCESS *im_access )
{
    int retval = -ENOTTY;
    uint32_t cmd_val = 0;
    extern FI_FPGA_DEV fi_fpga_dev;

//    dev_info(fi_fpga_dev.fpga_dev, "fi_fpga_image_read: enter");

    // Serialize access. Make sure to check if we're supposed to
    // block or not.
    if ((im_access->options & FI_IMA_NON_BLOCKING) == FI_IMA_NON_BLOCKING)
    {
	// Check to see if we can lock without blocking.
	if (!mutex_trylock(&im_lock))
	{
	    // Couldn't lock the mutex.
	    retval = -EACCES;
	    goto ir_bail;
	}
    }
    else
    {
	// Just lock the thing. Blocking is OK.
	mutex_lock(&im_lock);
    }

    // We now have access to the image memory. Setup the registers.
    fi_set_fpga_reg( &fi_fpga_dev, IM2PCIE_DMA_CONTROL, 0);
    fi_set_fpga_reg( &fi_fpga_dev, IM2PCIE_DMA_CONTROL, IM2PCIE_DMA_ENABLE);
    fi_set_fpga_reg( &fi_fpga_dev, IM2PCIE_IM_RD_DMA_START, im_access->im_address);
    fi_set_fpga_reg( &fi_fpga_dev, IM2PCIE_IM_RD_DMA_BLOCK_SIZE, im_access->im_block_size);
    fi_set_fpga_reg( &fi_fpga_dev, IM2PCIE_IM_RD_DMA_BURST_SIZE, im_access->im_burst_size);
    fi_set_fpga_reg( &fi_fpga_dev, IM2PCIE_IM_RD_DMA_BP_THRESHOLDS, im_access->im_bp_thresh);

    fi_set_fpga_reg( &fi_fpga_dev, IM2PCIE_PCIE_WR_DMA_START, im_access->pc_address);
    fi_set_fpga_reg( &fi_fpga_dev, IM2PCIE_PCIE_WR_DMA_BLOCK_SIZE, im_access->pc_block_size);
    fi_set_fpga_reg( &fi_fpga_dev, IM2PCIE_PCIE_WR_DMA_BURST_SIZE, im_access->pc_burst_size);
    fi_set_fpga_reg( &fi_fpga_dev, IM2PCIE_PCIE_WR_DMA_BURST_PERIOD, im_access->pc_burst_per);
    fi_set_fpga_reg( &fi_fpga_dev, IM2PCIE_PCIE_WR_DMA_BP_THRESHOLDS, im_access->pc_bp_thresh);
    fi_set_fpga_reg( &fi_fpga_dev, IM2PCIE_DMA_FRAME_DIMENSIONS, im_access->frame_dim);

    // Before we 'go', reset our event condition.
    frame_event_value = 0;

    // Tell the FPGA to start (maybe)...really, whatever is in the control word.
    cmd_val = IM2PCIE_DMA_ENABLE | im_access->control;
    fi_set_fpga_reg( &fi_fpga_dev, IM2PCIE_DMA_CONTROL, cmd_val);

    // Are we in blocking mode?
    if ((im_access->options & FI_IMA_NON_BLOCKING) != FI_IMA_NON_BLOCKING)
    {
	uint32_t wait_jiffies;
	uint32_t reg_value;
	long wait_status;
	int retries;
	const int RDIMAGE_TIMEOUT = 20;

	// Figure out how many jiffies we have to wait for. In general, we'd
	// like to get 20 frames/sec in live view. Of course when downloading
	// CAP files, this may well expire so just be prepared for that.
	wait_jiffies = HZ / 20;
	wait_status = 0;

	// Loop until we're done.
        for (retries = 0; retries < RDIMAGE_TIMEOUT; ++retries)
	{
	    // Check to see if frame_event_value got set by the interrupt handler.
	    if (frame_event_value == 0)
	    {
		// Enable the interrupt.
		fi_mod_fpga_reg(&fi_fpga_dev, PCIE_IRQ_MASK, 0, ~PCIEIRQ_IM2PCIE_DMA_DONE);

		// Wait for our event.
		wait_status = wait_event_interruptible_timeout(xfer_frame_q, (frame_event_value != 0), wait_jiffies);

		// Disable the interrupt.
		fi_mod_fpga_reg(&fi_fpga_dev, PCIE_IRQ_MASK, PCIEIRQ_IM2PCIE_DMA_DONE, -1);
	    }

            // Is the DMA done yet?
	    reg_value = fi_get_fpga_reg(&fi_fpga_dev, IM2PCIE_DMA_STATUS);
	    if ((reg_value & IM2PCIE_DMA_BUSY) == 0)
		break;

	    // If we're here, the DMA is still going. Sanity check to make sure
	    // the frame event value is still 0.
	    if (frame_event_value != 0)
	    {
		// Trying to think how this could be (without spelunking through FPGA code.)
		// Maybe (maybe?) the asynchronous nature of PCIe...the register space hasn't updated yet (the status register)
		// but we've got the interrupt. I wonder how long we wait on the next iteration before we see that DMA is not busy?
		//
		// As a note, I saw this when running live view with telemetry enabled.
		if (frame_event_value == PCIEIRQ_IM2PCIE_DMA_DONE)
		    printk("%s: DMA status is still busy, but frame_event_value is PCIEIRQ_IM2PCIE_DMA_DONE (we got the done interrupt)!", __FUNCTION__);
		else
		    printk("%s: DMA not done, but frame_event_value is 0x%08x!", __FUNCTION__, frame_event_value);
		frame_event_value = 0;
	    }

	    // Did we timeout?
	    if (wait_status == 0)
	    {
		printk("%s: [%d] TIMEOUT! IM2PCIE_DMA_STATUS=0x%08x\n", __FUNCTION__, retries, reg_value);
	    }
	    else
	    {
		// We didn't time-out, so backup the retry count.
		--retries;
	    }
	}

	// Did we exceed the number of retries?
	if (retries >= RDIMAGE_TIMEOUT)
	{
	    // Indicate an error.
	    printk("%s: RETRIES EXCEEDED [%d]! frame_event_value=0x%08x, IM2PCIE_DMA_STATUS=0x%08x\n", __FUNCTION__, retries, frame_event_value, reg_value);
	    reg_value = fi_get_fpga_reg(&fi_fpga_dev, IM2PCIE_DMA_CONTROL);
	    printk("%s: IM2PCIE_DMA_CONTROL=0x%08x\n", __FUNCTION__, reg_value);

	    retval = -ETIMEDOUT;
	}
	else
	{
	    // If we're here, then all is well.
	    retval = 0;
	}
    }

    // Release the lock.
    mutex_unlock(&im_lock);

ir_bail:
//    dev_info(fi_fpga_dev.fpga_dev, "fi_fpga_image_read: exit [%d]", retval);

    // Return our status.
    return retval;
}
EXPORT_SYMBOL(fi_fpga_image_read);

int fi_fpga_image_write( FI_IMAGE_ACCESS *im_access )
{
    int retval = -ENOTTY;
    extern FI_FPGA_DEV fi_fpga_dev;

//    dev_info(fi_fpga_dev.fpga_dev, "fi_fpga_image_write: enter");

    // Serialize access. Make sure to check if we're supposed to
    // block or not.
    if ((im_access->options & FI_IMA_NON_BLOCKING) == FI_IMA_NON_BLOCKING)
    {
	// Don't currently support non-blocking mode.
	dev_info(fi_fpga_dev.fpga_dev, "NON_BLOCKING mode not supported");
	retval = -EACCES;
	goto iw_bail;
    }
    else
    {
	// Just lock the thing. Blocking is OK.
	mutex_lock(&im_lock);
    }

    // We now have access to the image memory. Setup the registers.
    fi_set_fpga_reg( &fi_fpga_dev, PCIE2IM_DMA_CONTROL, 0);
    fi_set_fpga_reg( &fi_fpga_dev, PCIE2IM_DMA_CONTROL, PCIE2IM_DMA_ENABLE);

    fi_set_fpga_reg( &fi_fpga_dev, PCIE2IM_IM_WR_DMA_START, im_access->im_address);
    fi_set_fpga_reg( &fi_fpga_dev, PCIE2IM_IM_WR_DMA_BURST_SIZE, im_access->im_burst_size);
    fi_set_fpga_reg( &fi_fpga_dev, PCIE2IM_IM_WR_DMA_BP_THRESHOLDS, im_access->im_bp_thresh);

    fi_set_fpga_reg( &fi_fpga_dev, PCIE2IM_PCIE_RD_DMA_START, im_access->pc_address);
    fi_set_fpga_reg( &fi_fpga_dev, PCIE2IM_PCIE_RD_DMA_BLOCK_SIZE, im_access->pc_block_size);
    fi_set_fpga_reg( &fi_fpga_dev, PCIE2IM_PCIE_RD_DMA_BURST_SIZE, im_access->pc_burst_size);
    fi_set_fpga_reg( &fi_fpga_dev, PCIE2IM_PCIE_RD_DMA_BP_THRESHOLDS, im_access->pc_bp_thresh);

    // Before we 'go', reset our event condition.
    frame_event_value = 0;

    // Tell the FPGA to start (maybe)...really, whatever is in the control word.
    fi_set_fpga_reg( &fi_fpga_dev, PCIE2IM_DMA_CONTROL, PCIE2IM_DMA_ENABLE | im_access->control);

    // Are we in blocking mode?
    if ((im_access->options & FI_IMA_NON_BLOCKING) != FI_IMA_NON_BLOCKING)
    {
	int wait_status;
	uint32_t reg_value;
	uint32_t wait_jiffies;
	int retries;
	const int WRIMAGE_TIMEOUT = 10;

	// Figure out how many jiffies we have to wait for.
	wait_jiffies = HZ / 10;
	wait_status = 0;

	// Loop until we're done.
        for (retries = 0; retries < WRIMAGE_TIMEOUT; ++retries)
	{
	    // Check to see if frame_event_value got set by the interrupt handler.
	    if (frame_event_value == 0)
	    {
		// Enable the interrupt.
		fi_mod_fpga_reg(&fi_fpga_dev, PCIE_IRQ_MASK, 0, ~PCIEIRQ_PCIE2IM_DMA_DONE);

		// We're blocking, so wait for the operation to complete.
		wait_status = wait_event_interruptible_timeout(xfer_frame_q, (frame_event_value != 0), wait_jiffies);

		// Disable the interrupt.
		fi_mod_fpga_reg(&fi_fpga_dev, PCIE_IRQ_MASK, PCIEIRQ_PCIE2IM_DMA_DONE, -1);
	    }

	    // Is the DMA done yet?
	    reg_value = fi_get_fpga_reg(&fi_fpga_dev, PCIE2IM_DMA_STATUS);
	    if ((reg_value & PCIE2IM_DMA_BUSY) == 0)
		break;

	    // If we're here, the DMA is still going. Sanity check to make sure
	    // the frame event value is still 0.
	    if (frame_event_value != 0)
	    {
		printk("%s: DMA not done, but frame_event_value is 0x%08x!", __FUNCTION__, frame_event_value);
		frame_event_value = 0;
	    }

	    // Did we timeout?
	    if (wait_status == 0)
	    {
		printk("%s: [%d] TIMEOUT! PCIE2IM_DMA_STATUS=0x%08x\n", __FUNCTION__, retries, reg_value);
	    }
	    else
	    {
		// We didn't time-out, so backup the retry count.
		--retries;
	    }
	}

	if (retries >= WRIMAGE_TIMEOUT)
	{
	    // Indicate an error.
	    printk("%s: RETRIES EXCEEDED [%d]! frame_event_value=0x%08x, PCIE2IM_DMA_STATUS=0x%08x\n", __FUNCTION__, retries, frame_event_value, reg_value);
	    reg_value = fi_get_fpga_reg(&fi_fpga_dev, PCIE2IM_DMA_CONTROL);
	    printk("%s: PCIE2IM_DMA_CONTROL=0x%08x\n", __FUNCTION__, reg_value);
	    retval = -ETIMEDOUT;
	}
	else
	{
	    // If we're here, then all is well.
	    retval = 0;
	}
    }

    // Release the lock.
    mutex_unlock(&im_lock);

iw_bail:
//    dev_info(fi_fpga_dev.fpga_dev, "fi_fpga_image_write: exit [%d]", retval);

    // Return our status.
    return retval;
}
EXPORT_SYMBOL(fi_fpga_image_write);

