/*! 
 * @file fpga_io.c
 * @brief Register I/O operations for the FPGA device driver.
 *
 * Copyright (C) 2010,2015 Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------

#include <linux/device.h>
#include <linux/io.h>

#include "fi_pci.h"
#include "fi_fpga_dev.h"

/*! 
 * @brief Write a value to a FPGA register.
 * 
 * @param dev Pointer to our device structure.
 * @param reg_id The register we want to write to.
 * @param reg_value The value we want to write to the register.
 */
void fi_set_fpga_reg( FI_FPGA_DEV *dev, uint32_t reg_id, uint32_t reg_value )
{
    ulong	flags;
    uint32_t	*ioaddr;

    // Sanity check for a null device pointer.
    if ( dev == NULL )
	return;

    // Sanity check register ID value.
    if ((reg_id < FI_FPGA_BASE_ADDR) || (reg_id > FI_FPGA_MAX_ADDR))
    {
	/* If the caller dereferences the zero we return, we'll get a kernel oops. */
	dev_err( dev->fpga_dev, "%s: Bad reg_id 0x%08x!\n", __FUNCTION__, reg_id );
	return;
    }

    // Write the value to the register.
    ioaddr = fi_pci_get_base();
    if (ioaddr == NULL)
    {
	dev_err(dev->fpga_dev, "%s: No base address available!\n", __FUNCTION__);
	return;
    }

    // The register ID is a byte offset value. We're adding it 
    // to a 32-bit pointer, so we need to make the ID a word offset.
    reg_id >>= 2;

    // Perform the write.
    spin_lock_irqsave( &dev->reg_lock, flags );
    iowrite32( reg_value, ioaddr + reg_id);
    spin_unlock_irqrestore( &dev->reg_lock, flags );
}

/*! 
 * @brief Read a value from a FPGA register.
 * 
 * @param dev Pointer to our device structure.
 * @param reg_id The register we want to read from.
 * 
 * @return The value read from reg_id.
 */
uint32_t fi_get_fpga_reg( FI_FPGA_DEV *dev, uint32_t reg_id )
{
    uint32_t	ret_val;
    ulong	flags;
    uint32_t	*ioaddr;

    // Sanity check for a null device pointer.
    if ( dev == NULL )
	return reg_id;

    // Sanity check register ID value.
    if ((reg_id < FI_FPGA_BASE_ADDR) || (reg_id > FI_FPGA_MAX_ADDR))
    {
	/* If the caller dereferences the zero we return, we'll get a kernel oops. */
	dev_err( dev->fpga_dev, "%s: Bad reg_id 0x%08x!\n", __FUNCTION__, reg_id );
	return 0;
    }

    // Read the value from the register.
    ioaddr = fi_pci_get_base();
    if (ioaddr == NULL)
    {
	dev_err(dev->fpga_dev, "%s: No base address available!\n", __FUNCTION__);
	return 0;
    }

    // The register ID is a byte offset value. We're adding it 
    // to a 32-bit pointer, so we need to make the ID a word offset.
    reg_id >>= 2;

    spin_lock_irqsave( &dev->reg_lock, flags );
    ret_val = ioread32( ioaddr + reg_id );
    spin_unlock_irqrestore( &dev->reg_lock, flags );

    return ret_val;
}

/*!
 * @brief Peform a Read-Modify-Write operation on the FPGA.
 *
 * reg_value = ( reg_value & mask ) | value;
 *
 * @param reg_addr	Our FPGA register we'll be modifying.
 * @param value		Value to set into the register.
 * @param mask		Value to mask with the existing contents of the register.
 *
 * Performs
 *
 *	new_reg = ( old_reg & mask ) | value
 *
 * To set a bit
 *
 *	fpga_modify_reg( reg_name, BIT_NAME, -1 );
 *
 * To clear a bit
 *
 *	fpga_modify_reg( reg_name, 0, ~BIT_NAME );
 *
 * Example of changing a set of bits:
 *
 *	value =    ILC_BIT_DEPTH_8    | ILC_BIT_SEL_11_4;
 *	mask  = ~( ILC_BIT_DEPTH_MASK | ILC_BIT_SEL_MASK );
 *	fpga_modify_reg( reg_name, value, mask );
 *
 * @return true if everything went well, otherwise false.
 */
uint32_t fi_mod_fpga_reg( FI_FPGA_DEV *dev, uint32_t reg_id, uint32_t reg_value, uint32_t reg_mask )
{
    uint32_t	temp_value;

    // Sanity check for a null device pointer.
    if ( dev == NULL )
	return ((uint32_t)-EFAULT);

    // Make sure the register ID makes sense.
    if ( !FI_FPGA_REG_VALID( reg_id ))
    {
	dev_err( dev->fpga_dev, "Bad modify reg 0x%08x\n", reg_id );
	return ((uint32_t)-EFAULT);
    }

    // Grab the info from the FPGA.
    temp_value = fi_get_fpga_reg( dev, reg_id );

    // Mask, modify and write it back out.
    reg_value = ( temp_value & reg_mask ) | reg_value;

    // Update the register on the FPGA.
    fi_set_fpga_reg( dev, reg_id, reg_value );

    // All is good.
    return reg_value;
}

