/*! 
 * @file main.c
 * @brief Module entry point
 *
 * Copyright (C) 2010,2016 Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------

#include <linux/module.h>   	/* Needed by all modules */
#include <linux/kernel.h>	/* We're doing kernel development. */
#include <linux/device.h>
#include <linux/fs.h>		/* everything... */
#include <linux/init.h>		/* Needed for the macros */
#include <linux/io.h>

#include "fi_fpga_regs.h"
#include "fi_fpga_dev.h"

//-----------------------------------------------------------------------------
// Constant Definitions
//-----------------------------------------------------------------------------

// The maximum number of minor device nodes we'll support.
#define MAX_MINOR		1

//-----------------------------------------------------------------------------
// Variable Declarations
//-----------------------------------------------------------------------------

//! Our device structure.
FI_FPGA_DEV	fi_fpga_dev;

//-----------------------------------------------------------------------------
// Our parameters which can be set at load time.
//-----------------------------------------------------------------------------

int fi_fpga_verbose = 1;

module_param_named( verbose, fi_fpga_verbose, int, S_IRUGO | S_IWUSR );
MODULE_PARM_DESC( verbose, "Make the driver verbose" );

//! Our file operations structure.
struct file_operations fi_fpga_fops = {
    .owner =          THIS_MODULE,
    .unlocked_ioctl = fi_fpga_ioctl,
    .open =           fi_fpga_open,
    .release =        fi_fpga_release,
};

//-----------------------------------------------------------------------------
// Function Prototypes
//-----------------------------------------------------------------------------

/*!
 * @brief Used by the FPGA class, we're telling the filesystem the permissions
 * 	  to put on the /dev/fpga node.
 */
static int fi_fpga_uevent(struct device *dev, struct kobj_uevent_env *env)
{
    add_uevent_var(env, "DEVMODE=%#o", 0666);
    return 0;
}

/*! 
 * @brief Set up our char device.
 */
static int __init fi_fpga_setup_cdev(void)
{
    int   retval;

    printk("%s\n", __FUNCTION__);

    // Dynamically allocate our major number with only one minor number.
    retval = alloc_chrdev_region( &fi_fpga_dev.devno, 0, MAX_MINOR, MODULE_NAME );
    if ( retval < 0 )
	return retval;

    // Initialize our cdev structure.
    cdev_init( &fi_fpga_dev.fpga_cdev, &fi_fpga_fops );

    // Assign some of the fields.
    fi_fpga_dev.fpga_cdev.owner = THIS_MODULE;
    fi_fpga_dev.fpga_cdev.ops = &fi_fpga_fops;

    // Create a unique class because there's no one like us.
    fi_fpga_dev.fpga_class = class_create( THIS_MODULE, MODULE_NAME );
    if ( IS_ERR( fi_fpga_dev.fpga_class ))
    {
	retval = PTR_ERR( fi_fpga_dev.fpga_class );
	goto err1;
    }

    // Hook in our uevent handler so the /dev/fi_fpga node will have
    // the correct permissions.
    fi_fpga_dev.fpga_class->dev_uevent = fi_fpga_uevent;

    // Create the device for this class.
    fi_fpga_dev.fpga_dev = device_create( fi_fpga_dev.fpga_class, NULL, fi_fpga_dev.devno, NULL, MODULE_NAME );
    if ( IS_ERR( fi_fpga_dev.fpga_dev ))
    {
	retval = PTR_ERR( fi_fpga_dev.fpga_dev );
	goto err2;
    }

    // Add this device to the list.
    retval = cdev_add( &fi_fpga_dev.fpga_cdev, fi_fpga_dev.devno, MAX_MINOR );
    if ( retval < 0 )
    {
	cdev_del( &fi_fpga_dev.fpga_cdev );
	goto err3;
    }

    return 0;

err3:
    device_destroy( fi_fpga_dev.fpga_class, fi_fpga_dev.devno );
    fi_fpga_dev.fpga_dev = NULL;
err2:
    class_destroy( fi_fpga_dev.fpga_class );
    fi_fpga_dev.fpga_class = NULL;
err1:
    unregister_chrdev_region( fi_fpga_dev.devno, MAX_MINOR );

    // Return our status.
    return retval;
}

static int __init fi_fpga_mod_init(void)
{
    int rc = 0;

    printk(KERN_INFO "Fastec FPGA Driver: Welcome!\n");

    // Initialize our character device info.
    rc = fi_fpga_setup_cdev();
    if (rc < 0)
    {
	printk( KERN_ERR "%s - Unable to setup userland driver.\n", __FUNCTION__ );
	return rc;
    }

    // Initialize our hwmon device info.
    rc = fi_fpga_setup_hdev();
    if (rc < 0)
    {
	printk( KERN_ERR "%s - Unable to setup hwmon driver.\n", __FUNCTION__ );
	return rc;
    }

    // Initialize our locking control.
    spin_lock_init( &fi_fpga_dev.reg_lock );

    // Start at this number of available connections and count down
    // when something opens the device.
    atomic_set(&fi_fpga_dev.available, 1);

    return rc;
}

static void __exit fi_fpga_mod_exit(void)
{
    // Unregister the userland interface.
    cdev_del( &fi_fpga_dev.fpga_cdev );
    device_destroy( fi_fpga_dev.fpga_class, fi_fpga_dev.devno );
    class_destroy( fi_fpga_dev.fpga_class );
    unregister_chrdev_region( fi_fpga_dev.devno, MAX_MINOR );

    printk(KERN_INFO "Fastec FPGA Driver: Exiting...\n");
}

module_init(fi_fpga_mod_init);
module_exit(fi_fpga_mod_exit);

MODULE_AUTHOR("Fastec Imaging, Inc.");          /* Who wrote this. */
MODULE_DESCRIPTION("Fastec FPGA Driver");  	/* Helpful info.   */
MODULE_LICENSE("GPL");        			/* Get rid of the taint message. */

