/*! 
 * @file fi_pci_proc.c
 * @brief /proc functions for the Fastec PCI device driver.
 *
 * Copyright (C) 2016 Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------

#include <linux/module.h>   /* Needed by all modules */
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/sched.h>
#include <linux/seq_file.h>
#include <linux/slab.h>
#include <linux/unistd.h>
#include <linux/printk.h>
#include <linux/uaccess.h>
#include <asm/fpu/api.h>

#include "fi_pci.h"
#include "fi_pci_dev.h"

//-----------------------------------------------------------------------------
// Constant Definitions
//-----------------------------------------------------------------------------

// #define DO_IOREAD	1

#define MAX_INPUT_LEN		64
#define FI_IMAGE_MEM_BASE	(8L * 1024L * 1024L * 1024L)

//-----------------------------------------------------------------------------
// Variable Declarations
//-----------------------------------------------------------------------------

// What BAR do we use for I/O?
extern int fi_bar_select;

static struct proc_dir_entry *pci_dir, *regs_file, *lastread_file, *bar0read_file, *barsel_file;
static struct proc_dir_entry *barsel_len_file, *im_base_file, *imdump_file;
static void * __iomem	bar0;
static uint32_t		base_addr;
static uint64_t		image_mem_base;
static char		dump_width;
static uint64_t		dump_offset;

static char *usage = {
    "b <hex_base_addr> 		Change the base address.\n"
    "r <hex_addr> 		Read the 32-bit value at address.\n"
    "w <hex_addr> <hex_value>	Write value to address.\n"
};

//-----------------------------------------------------------------------------
// Function Prototypes
//-----------------------------------------------------------------------------

#ifndef DO_IOREAD
// We need a buffer to hold our data.
static  uint32_t barsel_buf[32];

/**
 * Perform an accelerated *aligned* read from WC
 * @dst: destination pointer
 * @src: source pointer
 * @len: how many bytes to copy
 *
 * __memcpy_ntdqa copies @len bytes from @src to @dst using
 * non-temporal instructions where available. Note that all arguments
 * (@src, @dst) must be aligned to 16 bytes and @len must be a multiple
 * of 16.
 */
static void __memcpy_ntdqa(void *dst, const void *src, unsigned long len)
{
    kernel_fpu_begin();

    len >>= 4;
    while (len >= 4)
    {
	asm("movntdqa   (%0), %%xmm0\n"
	    "movntdqa 16(%0), %%xmm1\n"
	    "movntdqa 32(%0), %%xmm2\n"
	    "movntdqa 48(%0), %%xmm3\n"
	    "movaps %%xmm0,   (%1)\n"
	    "movaps %%xmm1, 16(%1)\n"
	    "movaps %%xmm2, 32(%1)\n"
	    "movaps %%xmm3, 48(%1)\n"
	    :: "r" (src), "r" (dst) : "memory");
	src += 64;
	dst += 64;
	len -= 4;
    }
    while (len--)
    {
	asm("movntdqa (%0), %%xmm0\n"
	    "movaps %%xmm0, (%1)\n"
	    :: "r" (src), "r" (dst) : "memory");
	src += 16;
	dst += 16;
    }

    kernel_fpu_end();
}

static int proc_barsel_show( struct seq_file *m, void *v)
{
    int i;
    void * __iomem barN = fi_pci_priv_ptr->io_bar[fi_bar_select];

    if ((base_addr & 15) == 0)
    {
	// Read the data offset by base_addr.
	__memcpy_ntdqa(barsel_buf, barN+base_addr, 32*4);
    }
    else
    {
	// Read the data from address 0.
	__memcpy_ntdqa(barsel_buf, barN, 32*4);
    }

    // Remind the user which bar we are dumping.
    seq_printf(m, "BAR: %d\n", fi_bar_select);

    // Display 32 registers. Each register is on a 4-byte boundary,
    // which is why we're incrementing i by 4.
    for ( i = 0; i < 32; i++ )
    {
	if (( i & 0x3 ) == 0 )
	    seq_printf(m, "%08x: ", (base_addr + (i*4)));
	seq_printf(m, "%08x", barsel_buf[i]);

	if ((( i + 1 ) & 0x3 ) == 0 )
	    seq_printf( m, "\n" );
	else
	    seq_printf( m, " " );
    }

    // Add our final new line.
    seq_printf( m, "\n" );
    return 0;
}

static int proc_reg_show( struct seq_file *m, void *v)
{
    int i;

    if ((base_addr & 15) == 0)
    {
	// Read the data offset by base_addr.
	__memcpy_ntdqa(barsel_buf, (bar0 + base_addr), 32*4);
    }
    else
    {
	// Read the data from address 0.
	__memcpy_ntdqa(barsel_buf, bar0, 32*4);
    }

    // Display 32 registers. Each register is on a 4-byte boundary,
    // which is why we're incrementing i by 4.
    for ( i = 0; i < 32; i++ )
    {
	if (( i & 0x3 ) == 0 )
	    seq_printf(m, "%08x: ", (base_addr + (i*4)));
	seq_printf(m, "%08x", barsel_buf[i]);

	if ((( i + 1 ) & 0x3 ) == 0 )
	    seq_printf( m, "\n" );
	else
	    seq_printf( m, " " );
    }

    // Add our final new line.
    seq_printf( m, "\n" );
    return 0;
}
#endif /* !DO_IOREAD */

#ifdef DO_IOREAD
static int proc_reg_show( struct seq_file *m, void *v)
{
    int i;
    uint32_t read_val;

    // Display 32 registers. Each register is on a 4-byte boundary,
    // which is why we're incrementing i by 4.
    for ( i = 0; i < (32*4); i += 4 )
    {
	if (( i & 0xf ) == 0 )
	    seq_printf( m, "%08x: ", ( base_addr + i ));

	read_val = ioread32((u32 *)(bar0 + base_addr + i));
	seq_printf( m, "%08x", read_val );

	if ((( i + 4 ) & 0xF ) == 0 )
	    seq_printf( m, "\n" );
	else
	    seq_printf( m, " " );
    }

    // Add our final new line.
    seq_printf( m, "\n" );
    return 0;
}
#endif /* DO_IOREAD */

static int proc_reg_open(struct inode *inode, struct file *file)
{
    return single_open(file, proc_reg_show, NULL);
}

static ssize_t proc_reg_write( struct file *file, const char *buffer, size_t count, loff_t *offp )
{
    int len;
    char input_buf[ MAX_INPUT_LEN + 1 ];
    char	*end_ptr;
    uint32_t	user_addr;
    uint32_t	reg_value;
    
    if ( count > MAX_INPUT_LEN )
	len = MAX_INPUT_LEN;
    else
	len = count;

    if ( copy_from_user( input_buf, buffer, len ))
	return -EFAULT;

    input_buf[len+1] = '\0';
    switch ( input_buf[0] )
    {
    case 'b':
	base_addr = simple_strtoul( &input_buf[2], NULL, 16 );
	printk( "Setting base address to: 0x%x\n", base_addr );
	break;
    case 'h':
	printk( usage );
	break;
    case 'r':
	base_addr = simple_strtoul( &input_buf[2], NULL, 16 );
	reg_value = ioread32((u32 *)(bar0 + base_addr ));
	printk( "0x%08x = 0x%08x\n", base_addr, reg_value );
	break;
    case 'w':
	user_addr = simple_strtoul( &input_buf[2], &end_ptr, 16 );
	reg_value = simple_strtoul( end_ptr+1, NULL, 16 );
	iowrite32( reg_value, (u32 *)(bar0 + user_addr));
	break;
    default:
	printk( usage );
	break;
    }

    return len;
}

static const struct proc_ops proc_reg_fops = {
    .proc_open = proc_reg_open,
    .proc_read = seq_read,
    .proc_lseek = seq_lseek,
    .proc_release = single_release,
    .proc_write = proc_reg_write,
};

static int proc_last_show(struct seq_file *m, void *v)
{
    uint32_t reg_value;

    reg_value = ioread32((u32 *)(bar0 + base_addr));
    seq_printf( m, "0x%08x = 0x%08x\n", base_addr, reg_value );

    return 0;
}

static int proc_last_open(struct inode *inode, struct file *file)
{
    return single_open(file, proc_last_show, NULL);
}

static const struct proc_ops proc_lastread_fops = {
    .proc_open = proc_last_open,
    .proc_read = seq_read,
    .proc_lseek = seq_lseek,
    .proc_release = single_release,
};

static int proc_bar0_show(struct seq_file *m, void *v)
{
    seq_printf( m, "0x%p\n", bar0 );
    return 0;
}

static int proc_bar0_open(struct inode *inode, struct file *file)
{
    return single_open(file, proc_bar0_show, NULL);
}

static const struct proc_ops proc_bar0read_fops = {
    .proc_open = proc_bar0_open,
    .proc_read = seq_read,
    .proc_lseek = seq_lseek,
    .proc_release = single_release,
};

static int proc_barsel_len_show(struct seq_file *m, void *v)
{
    if (fi_pci_priv_ptr == NULL)
	seq_printf( m, "0\n" );
    else
	seq_printf( m, "%lu\n", fi_pci_priv_ptr->bar_length[fi_bar_select] );
    return 0;
}

static int proc_barsel_len_open(struct inode *inode, struct file *file)
{
    return single_open(file, proc_barsel_len_show, NULL);
}

static const struct proc_ops proc_barsel_len_fops = {
    .proc_open = proc_barsel_len_open,
    .proc_read = seq_read,
    .proc_lseek = seq_lseek,
    .proc_release = single_release,
};

static int proc_mem_show( struct seq_file *m, void *v)
{
#if 0
    int i;
    uint32_t read_val;
    uint8_t  *mem_ptr;

    // Make the image memory base to a pointer I can use.
    mem_ptr = (uint8_t *)ioremap(image_mem_base, 4096);

    // Display a page of memory.
    for ( i = 0; i < 4096; i += 16 )
    {
	int	j;

	seq_printf( m, "%llx: ", ( image_mem_base + i ));

	for (j = 0; j < 16; j++)
	    seq_printf( m, "%02x ", mem_ptr[i + j]);

	seq_printf( m, "\n" );
    }

    // Add our final new line.
    seq_printf( m, "\n" );
#endif

    // Display the image memory base address as we know it.
    seq_printf( m, "0x%llx\n", image_mem_base );
    return 0;
}

static int proc_mem_open(struct inode *inode, struct file *file)
{
    return single_open(file, proc_mem_show, NULL);
}

static ssize_t proc_mem_write( struct file *file, const char *buffer, size_t count, loff_t *offp )
{
    int len;
    char input_buf[ MAX_INPUT_LEN + 1 ];
    
    if ( count > MAX_INPUT_LEN )
	len = MAX_INPUT_LEN;
    else
	len = count;

    if ( copy_from_user( input_buf, buffer, len ))
	return -EFAULT;

    input_buf[len+1] = '\0';
    switch ( input_buf[0] )
    {
    case 'b':
	image_mem_base = simple_strtoull( &input_buf[2], NULL, 16 );
	printk( "Setting base address to: 0x%llx\n", image_mem_base );
	break;
    default:
	break;
    }

    return len;
}

static const struct proc_ops proc_mem_fops = {
    .proc_open = proc_mem_open,
    .proc_read = seq_read,
    .proc_lseek = seq_lseek,
    .proc_release = single_release,
    .proc_write = proc_mem_write,
};


/*!
 * @brief Return the image memory base address.
 */
uint64_t fi_pci_get_imbase( void )
{
    return image_mem_base;
}
EXPORT_SYMBOL(fi_pci_get_imbase);


static int proc_imdump_show( struct seq_file *m, void *v)
{
    int i;
    uint8_t  *mem_ptr;
    uint16_t *short_ptr;
    uint32_t *long_ptr;
    uint64_t dump_base;

    // Initialize our dump base address.
    dump_base = image_mem_base + dump_offset;

    // Make the image memory base to a pointer I can use.
    mem_ptr = (uint8_t *)ioremap(dump_base, 4096);

    // Display a page of memory.
    for ( i = 0; i < 4096; i += 16 )
    {
	int	j;

	seq_printf( m, "%llx: ", ( dump_base + i ));

	switch (dump_width)
	{
	case 'b':
	    for (j = 0; j < 16; j++)
		seq_printf( m, "%02x ", mem_ptr[i + j]);
	    break;
	case 's':
	    short_ptr = (uint16_t *)(&mem_ptr[i]);
	    for (j = 0; j < 8; j++)
		seq_printf( m, "%04x ", short_ptr[j]);
	    break;
	case 'l':
	    long_ptr = (uint32_t *)(&mem_ptr[i]);
	    for (j = 0; j < 4; j++)
		seq_printf( m, "%08x ", long_ptr[j]);
	    break;
	}

	seq_printf( m, "\n" );
    }

    // Add our final new line.
    seq_printf( m, "\n" );

    // Unmap the memory.
    iounmap(mem_ptr);

    return 0;
}

static int proc_imdump_open(struct inode *inode, struct file *file)
{
    return single_open(file, proc_imdump_show, NULL);
}

static ssize_t proc_imdump_write( struct file *file, const char *buffer, size_t count, loff_t *offp )
{
    int len;
    char input_buf[ MAX_INPUT_LEN + 1 ];
    
    if ( count > MAX_INPUT_LEN )
	len = MAX_INPUT_LEN;
    else
	len = count;

    if ( copy_from_user( input_buf, buffer, len ))
	return -EFAULT;

    input_buf[len+1] = '\0';
    switch ( input_buf[0] )
    {
    case 'b':
    case 's':
    case 'l':
	dump_width = input_buf[0];
	break;
    case 'o':
	dump_offset = simple_strtoull( &input_buf[2], NULL, 16 );
	printk( "Setting dump offset to: 0x%llx\n", dump_offset );
	break;
    case 'z':
	{
	    int i;
	    uint8_t  *mem_ptr;
	    uint64_t dump_base;

	    // Initialize our dump base address.
	    dump_base = image_mem_base + dump_offset;

	    // Make the image memory base to a pointer I can use.
	    mem_ptr = (uint8_t *)ioremap(dump_base, 4096);

	    // Zero the memory.
	    for ( i = 0; i < 4096; i++ )
		mem_ptr[i] = 0;

	    // Unmap the memory.
	    iounmap(mem_ptr);
	}
	break;
    case '?':
    case 'h':
    default:
	printk( "b, s, l: Display Byte, Short, or Long format\n");
	printk( "o <offset-in-hex>: Set the dump offset.\n" );
	printk( "z: Zero the dump memory.\n" );
	break;
    }

    return len;
}

static const struct proc_ops proc_imdump_fops = {
    .proc_open = proc_imdump_open,
    .proc_read = seq_read,
    .proc_lseek = seq_lseek,
    .proc_release = single_release,
    .proc_write = proc_imdump_write,
};

#ifdef DO_IOREAD
static int proc_barsel_show( struct seq_file *m, void *v)
{
    int i;
    uint32_t read_val;
    void * __iomem barN = fi_pci_priv_ptr->bar[fi_bar_select];

    seq_printf(m, "BAR: %d\n", fi_bar_select);

    // Loop through the bar, reading data.
    for ( i = 0; i < (32*4); i += 4 )
    {
	if (( i & 0xF ) == 0 )
	    seq_printf(m, "%08x: ", (base_addr + i));

	read_val = ioread32((u32 *)(barN + base_addr + i));
	seq_printf( m, "%08x", read_val );

	if ((( i + 4 ) & 0xF ) == 0 )
	    seq_printf( m, "\n" );
	else
	    seq_printf( m, " " );
    }

    // Add our final new line.
    seq_printf( m, "\n" );
    return 0;
}

#endif /* DO_IOREAD */

static int proc_barsel_open(struct inode *inode, struct file *file)
{
    return single_open(file, proc_barsel_show, NULL);
}

static const struct proc_ops proc_barsel_fops = {
    .proc_open = proc_barsel_open,
    .proc_read = seq_read,
    .proc_lseek = seq_lseek,
    .proc_release = single_release,
};

void __init fi_pci_proc_init(void * __iomem pci_bar0)
{
    pci_dir = proc_mkdir( MODULE_NAME, NULL );
    if ( pci_dir == NULL )
    {
	printk( "Unable to create pci_dir\n" );
	return;
    }

    regs_file = proc_create("regs", 0666, pci_dir, &proc_reg_fops);
    if (regs_file == NULL)
    {
	printk( "Unable to create regs_file\n" );
	remove_proc_entry(MODULE_NAME, NULL);
	return;
    }

    lastread_file = proc_create("last", 0444, pci_dir, &proc_lastread_fops );
    if (lastread_file == NULL)
    {
	printk( "Unable to create lastread_file\n" );
	remove_proc_entry("regs", pci_dir);
	remove_proc_entry(MODULE_NAME, NULL);
	return;
    }

    // Save the bar0 address.
    bar0 = pci_bar0;

    bar0read_file = proc_create("bar0", 0444, pci_dir, &proc_bar0read_fops );
    if (bar0read_file == NULL)
    {
	printk( "Unable to create bar0read_file\n" );
	remove_proc_entry("last", pci_dir);
	remove_proc_entry("regs", pci_dir);
	remove_proc_entry(MODULE_NAME, NULL);
	return;
    }

    // Initialize our default base address for image memory.
    image_mem_base = FI_IMAGE_MEM_BASE;

    im_base_file = proc_create("mem", 0666, pci_dir, &proc_mem_fops );
    if (im_base_file == NULL)
    {
	printk( "Unable to create im_base_file\n" );
	remove_proc_entry("bar0", pci_dir);
	remove_proc_entry("last", pci_dir);
	remove_proc_entry("regs", pci_dir);
	remove_proc_entry(MODULE_NAME, NULL);
	return;
    }

    // Default to a byte-wide image dump.
    dump_width = 'b';

    imdump_file = proc_create("imdump", 0666, pci_dir, &proc_imdump_fops );
    if (imdump_file == NULL)
    {
	printk( "Unable to create imdump_file\n" );
	remove_proc_entry("mem", pci_dir);
	remove_proc_entry("bar0", pci_dir);
	remove_proc_entry("last", pci_dir);
	remove_proc_entry("regs", pci_dir);
	remove_proc_entry(MODULE_NAME, NULL);
	return;
    }

    barsel_file = proc_create("bar_sel", 0666, pci_dir, &proc_barsel_fops );
    if (barsel_file == NULL)
    {
	printk( "Unable to create barsel_file\n" );
	remove_proc_entry("imdump", pci_dir);
	remove_proc_entry("mem", pci_dir);
	remove_proc_entry("bar0", pci_dir);
	remove_proc_entry("last", pci_dir);
	remove_proc_entry("regs", pci_dir);
	remove_proc_entry(MODULE_NAME, NULL);
	return;
    }

    barsel_len_file = proc_create("bar_sel_len", 0666, pci_dir, &proc_barsel_len_fops );
    if (barsel_len_file == NULL)
    {
	printk( "Unable to create barsel_len_file\n" );
	remove_proc_entry("bar_sel", pci_dir);
	remove_proc_entry("imdump", pci_dir);
	remove_proc_entry("mem", pci_dir);
	remove_proc_entry("bar0", pci_dir);
	remove_proc_entry("last", pci_dir);
	remove_proc_entry("regs", pci_dir);
	remove_proc_entry(MODULE_NAME, NULL);
	return;
    }

}

void __exit fi_pci_proc_exit( void )
{
    remove_proc_entry("bar_sel_len", pci_dir);
    remove_proc_entry("bar_sel", pci_dir);
    remove_proc_entry("imdump", pci_dir);
    remove_proc_entry("mem", pci_dir);
    remove_proc_entry("bar0", pci_dir);
    remove_proc_entry("last", pci_dir);
    remove_proc_entry("regs", pci_dir);
    remove_proc_entry(MODULE_NAME, NULL);
}


